from textual.app import App, ComposeResult
from textual.containers import Container
from textual.widgets import Input, Footer

from examples.custom_meta import ITEMS, get_items
from textual_autocomplete._autocomplete import AutoComplete, Dropdown, InputState


def completion_strategy(selected: str, input_state: InputState) -> InputState:
    return InputState(value="ham", cursor_position=1)


class Quadrant(App):
    def compose(self) -> ComposeResult:
        auto_completes = [
            AutoComplete(
                Input(
                    classes="search-box",
                    placeholder="Search for a UK city...",
                    id=f"input-{i}",
                ),
                Dropdown(items=ITEMS, classes=f"dropdown", id=f"dropdown-{i}"),
            )
            for i in range(1, 4)
        ]
        yield Container(
            *auto_completes,
            AutoComplete(
                Input(placeholder="Search for a UK city...", id="input-4"),
                Dropdown(
                    items=get_items,  # Using a callback to dynamically generate items
                    classes="dropdown",
                    id="dropdown-4",
                ),
                completion_strategy=completion_strategy,
            ),
            id="grid",
        )
        yield Footer()


app = Quadrant(css_path="styling.css")
if __name__ == "__main__":
    app.run()



from __future__ import annotations

from rich.color import Color
from rich.style import Style
from rich.text import Text
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Container
from textual.renderables._blend_colors import blend_colors
from textual.widgets import Input, Footer, Label

from textual_autocomplete._autocomplete import AutoComplete, DropdownItem, Dropdown, \
    InputState

INFO_TEXT = """\
Cities are ranked by population.
The left column shows the rank of the city.
The right column of the dropdown is coloured based on that population."""

DATA = [
    # ("London", "8,907,918"),  # Sorry London, u broke my naive colour-highlighting system
    ("Birmingham", "1,153,717"),
    ("Glasgow", "612,040"),
    ("Liverpool", "579,256"),
    ("Bristol", "571,922"),
    ("Manchester", "554,400"),
    ("Sheffield", "544,402"),
    ("Leeds", "503,388"),
    ("Edinburgh", "488,050"),
    ("Leicester", "470,965"),
    ("Coventry", "369,127"),
    ("Bradford", "361,046"),
    ("Cardiff", "350,558"),
    ("Belfast", "328,937"),
    ("Nottingham", "311,823"),
    ("Kingston upon Hull", "288,671"),
    ("Newcastle upon Tyne", "281,842"),
    ("Stoke-on-Trent", "277,051"),
    ("Southampton", "269,231"),
    ("Derby", "263,933"),
    ("Portsmouth", "248,479"),
    ("Brighton", "241,999"),
    ("Plymouth", "241,179"),
    ("Northampton", "229,815"),
    ("Reading", "229,274"),
    ("Luton", "222,907"),
    ("Wolverhampton", "218,255"),
    ("Bolton", "202,369"),
    ("Aberdeen", "200,680"),
    ("Bournemouth", "198,727"),
]

ITEMS = [
    DropdownItem(city, str(rank), population)
    for rank, (city, population) in enumerate(DATA, start=2)
]


def get_items(input_state: InputState) -> list[DropdownItem]:
    maximum_population = int(DATA[0][1].replace(",", ""))

    items = []

    # Let's create our dropdown items with custom colours in
    # the rightmost field (the higher the population, the more green)
    for rank, (city, population) in enumerate(DATA, start=2):
        ratio = float(population.replace(",", "")) / maximum_population
        color = blend_colors(Color.parse("#e86c4a"), Color.parse("#4ed43f"), ratio)
        items.append(
            DropdownItem(
                city,
                Text(str(rank), style="#a1a1a1"),
                Text(population, style=Style.from_color(color)),
            )
        )

    # Only keep cities that contain the Input value as a substring
    matches = [c for c in items if input_state.value.lower() in c.main.plain.lower()]
    # Favour items that start with the Input value, pull them to the top
    ordered = sorted(matches, key=lambda v: v.main.plain.startswith(input_state.value.lower()))

    return ordered


class CompletionExample(App):
    CSS_PATH = "custom_meta.css"

    BINDINGS = [Binding("ctrl+d", "toggle_dark", "Day/Night")]

    def compose(self) -> ComposeResult:
        yield Container(
            Label("Textual Autocomplete", id="lead-text"),
            AutoComplete(
                Input(id="search-box", placeholder="Search for a UK city..."),
                Dropdown(
                    items=get_items,  # Using a callback to dynamically generate items
                    id="my-dropdown",
                ),
            ),
            Label(INFO_TEXT, id="info-text"),
            id="search-container",
        )
        yield Footer()


app = CompletionExample()
if __name__ == "__main__":
    app.run()



"""Demonstrate a simple list of static items containing no metadata or fancy styling."""

from textual.app import App, ComposeResult
from textual.widgets import Input

from textual_autocomplete import AutoComplete, Dropdown, DropdownItem

ITEMS = [
    DropdownItem("Glasgow"),
    DropdownItem("Edinburgh"),
    DropdownItem("Aberdeen"),
    DropdownItem("Dundee"),
]


class Example01(App):

    CSS = """
    Dropdown { max-width: 32; }
    """

    def compose(self) -> ComposeResult:
        yield AutoComplete(Input(placeholder="type away..."), Dropdown(items=ITEMS))


app = Example01()

if __name__ == "__main__":
    app.run()



from __future__ import annotations

from dataclasses import dataclass
import functools
from operator import itemgetter
import types
from typing import (
    Callable,
    ClassVar,
    Iterator,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
    cast,
)
from rich.measure import Measurement
from rich.style import Style
from rich.text import Text, TextType
from textual import events
from textual.app import ComposeResult
from textual.binding import Binding
from textual.command import on
from textual.css.query import NoMatches
from textual_autocomplete.matcher import Matcher
from textual.geometry import Region, Spacing
from textual.widget import Widget
from textual.widgets import Input, TextArea, OptionList
from textual.widgets.option_list import Option
from textual.widgets.text_area import Location, Selection


@dataclass
class TargetState:
    text: str
    """The content in the target widget."""

    selection: Selection
    """The selection of the target widget."""


class SearchString(NamedTuple):
    start_location: int
    value: str


class InvalidTarget(Exception):
    """Raised if the target is invalid, i.e. not something which can
    be autocompleted."""


class DropdownItem(Option):
    def __init__(
        self,
        main: TextType,
        left_meta: TextType | None = None,
        # popup: TextType | None = None,
        id: str | None = None,
        disabled: bool = False,
    ) -> None:
        """A single option appearing in the autocompletion dropdown. Each option has up to 3 columns.
        Note that this is not a widget, it's simply a data structure for describing dropdown items.

        Args:
            left: The left column will often contain an icon/symbol, the main (middle)
                column contains the text that represents this option.
            main: The main text representing this option - this will be highlighted by default.
                In an IDE, the `main` (middle) column might contain the name of a function or method.
            search_string: The string that is being used for matching.
            highlight_ranges: Custom ranges to highlight. By default, the value is None,
                meaning textual-autocomplete will highlight substrings in the dropdown.
                That is, if the value you've typed into the Input is a substring of the candidates
                `main` attribute, then that substring will be highlighted. If you supply your own
                implementation of `items` which uses a more complex process to decide what to
                display in the dropdown, then you can customise the highlighting of the returned
                candidates by supplying index ranges to highlight.
        """
        self.main = Text(main, no_wrap=True) if isinstance(main, str) else main
        self.left_meta = (
            Text(left_meta, no_wrap=True, style="dim")
            if isinstance(left_meta, str)
            else left_meta
        )
        # self.popup = (
        #     Text(popup, no_wrap=True, style="dim") if isinstance(popup, str) else popup
        # )
        left = self.left_meta
        prompt = self.main
        if left:
            prompt = Text.assemble(left, " ", self.main)

        super().__init__(prompt, id, disabled)


class AutoCompleteList(OptionList):
    def get_content_width(self, container: events.Size, viewport: events.Size) -> int:
        """Get maximum width of options."""
        console = self.app.console
        options = console.options
        max_width = max(
            (
                Measurement.get(console, options, option.prompt).maximum
                for option in self._options
            ),
            default=1,
        )
        max_width += self.scrollbar_size_vertical
        return max_width


MatcherFactoryType = Callable[[str, Optional[Style], bool], Matcher]

TargetType = TypeVar("TargetType", bound=Union[Input, TextArea])


class AutoComplete(Widget):
    BINDINGS = [
        Binding("escape", "hide", "Hide dropdown", show=False),
    ]

    DEFAULT_CSS = """\
    AutoComplete {
        layer: textual-autocomplete;
        height: auto;
        width: auto;
        max-height: 12;
        display: none;
        background: $surface-lighten-1;

        & AutoCompleteList {
            width: auto;
            height: auto;
            border: none;
            padding: 0;
            margin: 0;
            scrollbar-size-vertical: 1;
            scrollbar-size-horizontal: 0;
            text-wrap: nowrap;
            &:focus {
                border: none;
                padding: 0;
                margin: 0;
            }
            & > .option-list--option-highlighted, & > .option-list--option-hover-highlighted {
                color: $text;
                background: $surface-lighten-3 60%;
            }
            
        }

        & .autocomplete--highlight-match {
            color: $text;
            background: $primary-lighten-1;
        }
    }
    """

    COMPONENT_CLASSES: ClassVar[set[str]] = {
        "autocomplete--highlight-match",
    }

    def __init__(
        self,
        target: Input | TextArea | str,
        candidates: list[DropdownItem] | Callable[[TargetState], list[DropdownItem]],
        matcher_factory: MatcherFactoryType | None = None,
        completion_strategy: (
            Callable[[str, TargetState], TargetState | None] | None
        ) = None,
        search_string: Callable[[TargetState], str] | None = None,
        prevent_default_enter: bool = True,
        prevent_default_tab: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self._target = target
        """An Input instance, TextArea instance, or a selector string used to query an Input/TextArea instance.
        
        Must be on the same screen as the dropdown instance."""

        self.completion_strategy = completion_strategy
        """A function which modifies the state of the target widget 
        to perform the completion.
        If None, the default behavior will be used.
        """

        self.candidates = candidates
        """The candidates to match on, or a function which returns the candidates to match on."""

        self.matcher_factory = Matcher if matcher_factory is None else matcher_factory
        """A factory function that returns the Matcher to use for matching and highlighting."""

        self.search_string = search_string
        """A function that returns the search string to match on.
        
        This is the string that will be passed into the matcher.

        If None, the default the default behavior will be used.

        For Input widgets, the default behavior is to use the entire value 
        as the search string.

        For TextArea widgets, the default behavior is to use the text before the cursor 
        as the search string, up until the last whitespace.
        """

        self.prevent_default_enter = prevent_default_enter
        """Prevent the default enter behavior."""

        self.prevent_default_tab = prevent_default_tab
        """Prevent the default tab behavior."""

        self.last_action_was_completion = False
        """Used to filter duplicate performing an action twice on a character insertion.
        An insertion/deletion moves the cursor and creates a "changed" event, so we end up
        with two events for the same action.
        """

        self._target_state = TargetState("", Selection.cursor((0, 0)))
        """Cached state of the target Input/TextArea."""

        self._patch_on_event()

    def compose(self) -> ComposeResult:
        option_list = AutoCompleteList()
        option_list.can_focus = False
        yield option_list

    def on_mount(self) -> None:
        # Subscribe to the target widget's reactive attributes.
        self.target.message_signal.subscribe(self, self._listen_to_messages)  # type: ignore
        self.screen.screen_layout_refresh_signal.subscribe(  # type: ignore
            self,
            lambda _event: self._align_to_target(),  # type: ignore
        )
        self._subscribe_to_target()
        self._handle_target_update()

    def _patch_on_event(self) -> None:
        target = self.target
        if isinstance(target, TextArea):
            original_on_event = target.on_event

            @functools.wraps(original_on_event)
            async def on_event(text_area: TextArea, event: events.Event) -> None:
                if (
                    isinstance(event, events.Key)
                    and event.key in {"tab", "enter"}
                    and self.styles.display != "none"
                ):
                    self._complete(self.option_list.highlighted or 0)
                    return
                return await original_on_event(event)

            bound_on_event = types.MethodType(on_event, target)
            setattr(target, "on_event", bound_on_event)

    def _listen_to_messages(self, event: events.Event) -> None:
        """Listen to some events of the target widget."""

        try:
            option_list = self.option_list
        except NoMatches:
            # This can happen if the event is an Unmount event
            # during application shutdown.
            return

        if isinstance(event, events.Key) and option_list.option_count:
            displayed = self.display
            highlighted = option_list.highlighted or 0
            if event.key == "down":
                # Check if there's only one item and it matches the search string
                if option_list.option_count == 1:
                    search_string = self.get_search_string(self._get_target_state())
                    first_option = option_list.get_option_at_index(0).prompt
                    text_from_option = (
                        first_option.plain
                        if isinstance(first_option, Text)
                        else first_option
                    )
                    if text_from_option == search_string:
                        # Don't prevent default behavior in this case
                        return

                event.stop()
                event.prevent_default()
                # If you press `down` while in an Input and the autocomplete is currently
                # hidden, then we should show the dropdown.
                if isinstance(self.target, Input):
                    if not displayed:
                        self.display = True
                        highlighted = 0
                    else:
                        highlighted = (highlighted + 1) % option_list.option_count
                else:
                    if displayed:
                        highlighted = (highlighted + 1) % option_list.option_count

                option_list.highlighted = highlighted

            elif event.key == "up":
                if displayed:
                    event.stop()
                    event.prevent_default()
                    highlighted = (highlighted - 1) % option_list.option_count
                    option_list.highlighted = highlighted
            elif event.key == "enter":
                if self.prevent_default_enter and displayed:
                    event.stop()
                    event.prevent_default()
                self._complete(option_index=highlighted)
            elif event.key == "tab":
                if self.prevent_default_tab and displayed:
                    event.stop()
                    event.prevent_default()
                self._complete(option_index=highlighted)
            elif event.key == "escape":
                self.action_hide()

        if isinstance(event, (Input.Changed, TextArea.Changed)):
            self._handle_target_update()

    def action_hide(self) -> None:
        self.styles.display = "none"

    def action_show(self) -> None:
        self.styles.display = "block"

    def _complete(self, option_index: int) -> None:
        """Do the completion (i.e. insert the selected item into the target input/textarea).

        This is when the user highlights an option in the dropdown and presses tab or enter.
        """
        if not self.display or self.option_list.option_count == 0:
            return

        target = self.target
        completion_strategy = self.completion_strategy
        option_list = self.option_list
        highlighted = option_index
        option = cast(DropdownItem, option_list.get_option_at_index(highlighted))
        highlighted_value = option.main.plain
        if isinstance(target, Input):
            if completion_strategy is None:
                target.value = ""
                target.insert_text_at_cursor(highlighted_value)
            elif callable(completion_strategy):
                completion_strategy(
                    highlighted_value,
                    self._get_target_state(),
                )
        else:  # elif isinstance(target, TextArea):
            if completion_strategy is None:
                replacement_range = self.get_text_area_word_bounds_before_cursor(target)
                target.replace(highlighted_value, *replacement_range)
            elif callable(completion_strategy):
                completion_strategy(
                    highlighted_value,
                    self._get_target_state(),
                )

        # Set a flag indicating that the last action that was performed
        # was a completion. This is so that when the target posts a Changed message
        # as a result of this completion, we can opt to ignore it in `handle_target_updated`
        self.last_action_was_completion = True
        self.action_hide()

    def yield_characters_before_cursor(
        self, target: TextArea
    ) -> Iterator[tuple[str, Location]]:
        cursor_location = target.cursor_location

        cursor_row, column = cursor_location
        start = (cursor_row, 0)
        text = target.get_text_range(start=start, end=cursor_location)
        for char in reversed(text):
            column -= 1
            yield char, (cursor_row, column)

    def get_text_area_word_bounds_before_cursor(
        self, target: TextArea
    ) -> tuple[Location, Location]:
        """Get the bounds of the word before the cursor in a TextArea.

        A word is defined as a sequence of alphanumeric characters or underscores,
        bounded by the start of the line, a space, or a non-alphanumeric character.

        Returns:
            A tuple containing the start and end positions of the word as (row, column) tuples.
        """
        cursor_location = target.cursor_location
        for char, (row, column) in self.yield_characters_before_cursor(target):
            if not char.isalnum() and char not in "$_-":
                return (row, column + 1), cursor_location
            elif column == 0:
                return (row, column), cursor_location

        return cursor_location, cursor_location

    @property
    def target(self) -> Input | TextArea:
        """The resolved target widget."""
        if isinstance(self._target, (Input, TextArea)):
            return self._target
        else:
            target = self.screen.query_one(self._target)
            assert isinstance(target, (Input, TextArea))
            return target

    def _subscribe_to_target(self) -> None:
        """Attempt to subscribe to the target widget, if it's available."""
        target = self.target
        self.watch(target, "has_focus", self._handle_focus_change)
        if isinstance(target, Input):
            self.watch(target, "cursor_position", self._align_to_target)
        else:
            self.watch(target, "selection", self._align_to_target)

    def _handle_target_message(self, message: events.Event) -> None:
        if isinstance(message, (Input.Changed, TextArea.Changed)):
            self._handle_target_update()

    def _align_to_target(self) -> None:
        cursor_x, cursor_y = self.target.cursor_screen_offset
        dropdown = self.query_one(OptionList)
        width, height = dropdown.size
        x, y, _width, _height = Region(
            cursor_x,
            cursor_y + 1,
            width,
            height,
        ).constrain("inside", "none", Spacing.all(2), self.screen.region)
        self.styles.offset = max(x - 1, 0), y

    def _get_target_state(self) -> TargetState:
        """Get the state of the target widget."""
        target = self.target
        if isinstance(target, Input):
            return TargetState(
                text=target.value,
                selection=Selection.cursor((0, target.cursor_position)),  # type: ignore
            )
        else:
            return TargetState(
                text=target.text,
                selection=target.selection,
            )

    def _handle_focus_change(self, has_focus: bool) -> None:
        """Called when the focus of the target widget changes."""
        if not has_focus:
            self.action_hide()
        else:
            search_string = self.get_search_string(self._target_state)
            self._rebuild_options(self._target_state, search_string)

    def _handle_target_update(self) -> None:
        """Called when the state (text or selection) of the target is updated.

        Here we align the dropdown to the target, determine if it should be visible,
        and rebuild the options in it.
        """
        self._target_state = self._get_target_state()
        search_string = self.get_search_string(self._target_state)

        # Determine visibility after the user makes a change in the
        # target widget (e.g. typing in a character in the Input).
        self._rebuild_options(self._target_state, search_string)
        self._align_to_target()

        if self.should_show_dropdown(search_string):
            self.action_show()
        else:
            self.action_hide()

        # We've rebuilt the options based on the latest change,
        # however, if the user made that change via selecting a completion
        # from the dropdown, then we always want to hide the dropdown.
        if self.last_action_was_completion:
            self.last_action_was_completion = False
            self.action_hide()
            return

    def should_show_dropdown(self, search_string: str) -> bool:
        """
        Determine whether to show or hide the dropdown based on the current state.

        This method can be overridden to customize the visibility behavior.

        Args:
            search_string: The current search string.

        Returns:
            bool: True if the dropdown should be shown, False otherwise.
        """
        option_list = self.option_list
        option_count = option_list.option_count

        if len(search_string) == 0 or option_count == 0:
            return False
        elif option_count == 1:
            first_option = option_list.get_option_at_index(0).prompt
            text_from_option = (
                first_option.plain if isinstance(first_option, Text) else first_option
            )
            return text_from_option != search_string
        else:
            return True

    def _rebuild_options(self, target_state: TargetState, search_string: str) -> None:
        """Rebuild the options in the dropdown.

        Args:
            target_state: The state of the target widget.
        """
        option_list = self.option_list
        option_list.clear_options()
        if self.target.has_focus:
            matches = self._compute_matches(target_state, search_string)
            if matches:
                option_list.add_options(matches)
                option_list.highlighted = 0

    def get_search_string(self, target_state: TargetState) -> str:
        """This value will be passed to the matcher.

        This could be, for example, the text in the target widget, or a substring of that text.

        For Input widgets the default is to use the text in the input, and for TextArea widgets
        the default is to use the text in the TextArea before the cursor up to the most recent
        non-alphanumeric character.

        Subclassing AutoComplete to create a custom `get_search_string` method is a way to
        customise the behaviour of the autocomplete dropdown.

        Returns:
            The search string that will be used to filter the dropdown options.
        """
        if self.search_string is not None:
            search_string = self.search_string(target_state)
            return search_string

        if isinstance(self.target, Input):
            return target_state.text
        else:
            start, end = self.get_text_area_word_bounds_before_cursor(self.target)
            search_string = self.target.get_text_range(start, end)
            return search_string

    def _compute_matches(
        self, target_state: TargetState, search_string: str
    ) -> list[DropdownItem]:
        """Compute the matches based on the target state.

        Args:
            target_state: The state of the target widget.

        Returns:
            The matches to display in the dropdown.
        """

        # If items is a callable, then it's a factory function that returns the candidates.
        # Otherwise, it's a list of candidates.
        candidates = self.get_candidates(target_state)
        return self.get_matches(target_state, candidates, search_string)

    def get_candidates(self, target_state: TargetState) -> list[DropdownItem]:
        """Get the candidates to match against."""
        candidates = self.candidates
        return candidates(target_state) if callable(candidates) else candidates

    def get_matches(
        self,
        target_state: TargetState,
        candidates: list[DropdownItem],
        search_string: str,
    ) -> list[DropdownItem]:
        """Given the state of the target widget, return the DropdownItems
        which match the query string and should be appear in the dropdown.

        Args:
            target_state: The state of the target widget.
            candidates: The candidates to match against.

        Returns:
            The matches to display in the dropdown.
        """
        if not search_string:
            return candidates

        match_style = self.get_component_rich_style("autocomplete--highlight-match")
        matcher = self.matcher_factory(search_string, match_style, False)

        matches_and_scores: list[tuple[DropdownItem, float]] = []
        append_score = matches_and_scores.append
        get_score = matcher.match
        get_highlighted = matcher.highlight

        for candidate in candidates:
            candidate_string = candidate.main.plain
            if (score := get_score(candidate_string)) > 0:
                highlighted_text = get_highlighted(candidate_string)
                highlighted_item = DropdownItem(
                    main=highlighted_text,
                    left_meta=candidate.left_meta,
                    id=candidate.id,
                    disabled=candidate.disabled,
                )
                append_score((highlighted_item, score))

        matches_and_scores.sort(key=itemgetter(1), reverse=True)
        matches = [match for match, _ in matches_and_scores]
        return matches

    @property
    def option_list(self) -> AutoCompleteList:
        return self.query_one(AutoCompleteList)

    @on(OptionList.OptionSelected, "AutoCompleteList")
    def _apply_completion(self, event: OptionList.OptionSelected) -> None:
        self._complete(event.option_index)



from __future__ import annotations
from functools import lru_cache

import os
import re
from typing import Callable, Literal

from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Vertical, VerticalScroll
from textual.widgets import Footer, Input, TextArea
from textual.widgets.text_area import Selection

from textual_autocomplete import (
    AutoComplete,
    DropdownItem,
    MatcherFactoryType,
    TargetState,
)

from rich.text import Text

from dataclasses import dataclass


@dataclass
class Header:
    name: str
    example: str
    description: str


headers = [
    {
        "section": "Authentication",
        "name": "WWW-Authenticate",
        "description": "Defines the authentication method that should be used to access a resource.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/WWW-Authenticate",
    },
    {
        "section": "Authentication",
        "name": "Authorization",
        "description": "Contains the credentials to authenticate a user-agent with a server.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Authorization",
    },
    {
        "section": "Authentication",
        "name": "Proxy-Authenticate",
        "description": "Defines the authentication method that should be used to access a resource behind a proxy server.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Proxy-Authenticate",
    },
    {
        "section": "Authentication",
        "name": "Proxy-Authorization",
        "description": "Contains the credentials to authenticate a user agent with a proxy server.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Proxy-Authorization",
    },
    {
        "section": "Caching",
        "name": "Age",
        "description": "The time, in seconds, that the object has been in a proxy cache.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Age",
    },
    {
        "section": "Caching",
        "name": "Cache-Control",
        "description": "Directives for caching mechanisms in both requests and responses.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Cache-Control",
    },
    {
        "section": "Caching",
        "name": "Clear-Site-Data",
        "description": "Clears browsing data (e.g. cookies, storage, cache) associated with the requesting website.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Clear-Site-Data",
    },
    {
        "section": "Caching",
        "name": "Expires",
        "description": "The date/time after which the response is considered stale.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Expires",
    },
    {
        "section": "Caching",
        "name": "No-Vary-Search",
        "description": "Specifies a set of rules that define how a URL's query parameters will affect cache matching. These rules dictate whether the same URL with different URL parameters should be saved as separate browser cache entries.",
        "experimental": True,
        "url": "/en-US/docs/Web/HTTP/Headers/No-Vary-Search",
    },
    {
        "section": "Conditionals",
        "name": "Last-Modified",
        "description": "The last modification date of the resource, used to compare several versions of the same resource. It is less accurate than ETag, but easier to calculate in some environments. Conditional requests using If-Modified-Since and If-Unmodified-Since use this value to change the behavior of the request.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Last-Modified",
    },
    {
        "section": "Conditionals",
        "name": "ETag",
        "description": "A unique string identifying the version of the resource. Conditional requests using If-Match and If-None-Match use this value to change the behavior of the request.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/ETag",
    },
    {
        "section": "Conditionals",
        "name": "If-Match",
        "description": "Makes the request conditional, and applies the method only if the stored resource matches one of the given ETags.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/If-Match",
    },
    {
        "section": "Conditionals",
        "name": "If-None-Match",
        "description": "Makes the request conditional, and applies the method only if the stored resource doesn't match any of the given ETags. This is used to update caches (for safe requests), or to prevent uploading a new resource when one already exists.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/If-None-Match",
    },
    {
        "section": "Conditionals",
        "name": "If-Modified-Since",
        "description": "Makes the request conditional, and expects the resource to be transmitted only if it has been modified after the given date. This is used to transmit data only when the cache is out of date.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/If-Modified-Since",
    },
    {
        "section": "Conditionals",
        "name": "If-Unmodified-Since",
        "description": "Makes the request conditional, and expects the resource to be transmitted only if it has not been modified after the given date. This ensures the coherence of a new fragment of a specific range with previous ones, or to implement an optimistic concurrency control system when modifying existing documents.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/If-Unmodified-Since",
    },
    {
        "section": "Conditionals",
        "name": "Vary",
        "description": "Determines how to match request headers to decide whether a cached response can be used rather than requesting a fresh one from the origin server.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Vary",
    },
    {
        "section": "Connection management",
        "name": "Connection",
        "description": "Controls whether the network connection stays open after the current transaction finishes.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Connection",
    },
    {
        "section": "Connection management",
        "name": "Keep-Alive",
        "description": "Controls how long a persistent connection should stay open.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Keep-Alive",
    },
    {
        "section": "Content negotiation",
        "name": "Accept",
        "description": "Informs the server about the types of data that can be sent back.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Accept",
    },
    {
        "section": "Content negotiation",
        "name": "Accept-Encoding",
        "description": "The encoding algorithm, usually a compression algorithm, that can be used on the resource sent back.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Accept-Encoding",
    },
    {
        "section": "Content negotiation",
        "name": "Accept-Language",
        "description": "Informs the server about the human language the server is expected to send back. This is a hint and is not necessarily under the full control of the user: the server should always pay attention not to override an explicit user choice (like selecting a language from a dropdown).",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Accept-Language",
    },
    {
        "section": "Controls",
        "name": "Expect",
        "description": "Indicates expectations that need to be fulfilled by the server to properly handle the request.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Expect",
    },
    {
        "section": "Controls",
        "name": "Max-Forwards",
        "description": "When using TRACE, indicates the maximum number of hops the request can do before being reflected to the sender.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Max-Forwards",
    },
    {
        "section": "Cookies",
        "name": "Cookie",
        "description": "Contains stored HTTP cookies previously sent by the server with the Set-Cookie header.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Cookie",
    },
    {
        "section": "Cookies",
        "name": "Set-Cookie",
        "description": "Send cookies from the server to the user-agent.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Set-Cookie",
    },
    {
        "section": "CORS",
        "name": "Access-Control-Allow-Credentials",
        "description": "Indicates whether the response to the request can be exposed when the credentials flag is true.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Credentials",
    },
    {
        "section": "CORS",
        "name": "Access-Control-Allow-Headers",
        "description": "Used in response to a preflight request to indicate which HTTP headers can be used when making the actual request.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Headers",
    },
    {
        "section": "CORS",
        "name": "Access-Control-Allow-Methods",
        "description": "Specifies the methods allowed when accessing the resource in response to a preflight request.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Methods",
    },
    {
        "section": "CORS",
        "name": "Access-Control-Allow-Origin",
        "description": "Indicates whether the response can be shared.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Origin",
    },
    {
        "section": "CORS",
        "name": "Access-Control-Expose-Headers",
        "description": "Indicates which headers can be exposed as part of the response by listing their names.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Access-Control-Expose-Headers",
    },
    {
        "section": "CORS",
        "name": "Access-Control-Max-Age",
        "description": "Indicates how long the results of a preflight request can be cached.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Access-Control-Max-Age",
    },
    {
        "section": "CORS",
        "name": "Access-Control-Request-Headers",
        "description": "Used when issuing a preflight request to let the server know which HTTP headers will be used when the actual request is made.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Access-Control-Request-Headers",
    },
    {
        "section": "CORS",
        "name": "Access-Control-Request-Method",
        "description": "Used when issuing a preflight request to let the server know which HTTP method will be used when the actual request is made.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Access-Control-Request-Method",
    },
    {
        "section": "CORS",
        "name": "Origin",
        "description": "Indicates where a fetch originates from.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Origin",
    },
    {
        "section": "CORS",
        "name": "Timing-Allow-Origin",
        "description": "Specifies origins that are allowed to see values of attributes retrieved via features of the Resource Timing API, which would otherwise be reported as zero due to cross-origin restrictions.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Timing-Allow-Origin",
    },
    {
        "section": "Downloads",
        "name": "Content-Disposition",
        "description": 'Indicates if the resource transmitted should be displayed inline (default behavior without the header), or if it should be handled like a download and the browser should present a "Save As" dialog.',
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Content-Disposition",
    },
    {
        "section": "Message body information",
        "name": "Content-Length",
        "description": "The size of the resource, in decimal number of bytes.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Content-Length",
    },
    {
        "section": "Message body information",
        "name": "Content-Type",
        "description": "Indicates the media type of the resource.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Content-Type",
    },
    {
        "section": "Message body information",
        "name": "Content-Encoding",
        "description": "Used to specify the compression algorithm.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Content-Encoding",
    },
    {
        "section": "Message body information",
        "name": "Content-Language",
        "description": "Describes the human language(s) intended for the audience, so that it allows a user to differentiate according to the users' own preferred language.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Content-Language",
    },
    {
        "section": "Message body information",
        "name": "Content-Location",
        "description": "Indicates an alternate location for the returned data.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Content-Location",
    },
    {
        "section": "Proxies",
        "name": "Forwarded",
        "description": "Contains information from the client-facing side of proxy servers that is altered or lost when a proxy is involved in the path of the request.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Forwarded",
    },
    {
        "section": "Proxies",
        "name": "Via",
        "description": "Added by proxies, both forward and reverse proxies, and can appear in the request headers and the response headers.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Via",
    },
    {
        "section": "Redirects",
        "name": "Location",
        "description": "Indicates the URL to redirect a page to.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Location",
    },
    {
        "section": "Request context",
        "name": "From",
        "description": "Contains an Internet email address for a human user who controls the requesting user agent.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/From",
    },
    {
        "section": "Request context",
        "name": "Host",
        "description": "Specifies the domain name of the server (for virtual hosting), and (optionally) the TCP port number on which the server is listening.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Host",
    },
    {
        "section": "Request context",
        "name": "Referer",
        "description": "The address of the previous web page from which a link to the currently requested page was followed.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Referer",
    },
    {
        "section": "Request context",
        "name": "Referrer-Policy",
        "description": "Governs which referrer information sent in the Referer header should be included with requests made.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Referrer-Policy",
    },
    {
        "section": "Request context",
        "name": "User-Agent",
        "description": "Contains a characteristic string that allows the network protocol peers to identify the application type, operating system, software vendor or software version of the requesting software user agent.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/User-Agent",
    },
    {
        "section": "Response context",
        "name": "Allow",
        "description": "Lists the set of HTTP request methods supported by a resource.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Allow",
    },
    {
        "section": "Response context",
        "name": "Server",
        "description": "Contains information about the software used by the origin server to handle the request.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Server",
    },
    {
        "section": "Range requests",
        "name": "Accept-Ranges",
        "description": "Indicates if the server supports range requests, and if so in which unit the range can be expressed.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Accept-Ranges",
    },
    {
        "section": "Range requests",
        "name": "Range",
        "description": "Indicates the part of a document that the server should return.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Range",
    },
    {
        "section": "Range requests",
        "name": "If-Range",
        "description": "Creates a conditional range request that is only fulfilled if the given etag or date matches the remote resource. Used to prevent downloading two ranges from incompatible version of the resource.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/If-Range",
    },
    {
        "section": "Range requests",
        "name": "Content-Range",
        "description": "Indicates where in a full body message a partial message belongs.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Content-Range",
    },
    {
        "section": "Security",
        "name": "Cross-Origin-Embedder-Policy",
        "description": "Allows a server to declare an embedder policy for a given document.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Cross-Origin-Embedder-Policy",
    },
    {
        "section": "Security",
        "name": "Cross-Origin-Opener-Policy",
        "description": "Prevents other domains from opening/controlling a window.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Cross-Origin-Opener-Policy",
    },
    {
        "section": "Security",
        "name": "Cross-Origin-Resource-Policy",
        "description": "Prevents other domains from reading the response of the resources to which this header is applied. See also CORP explainer article.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Cross-Origin-Resource-Policy",
    },
    {
        "section": "Security",
        "name": "Content-Security-Policy",
        "description": "Controls resources the user agent is allowed to load for a given page.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Content-Security-Policy",
    },
    {
        "section": "Security",
        "name": "Content-Security-Policy-Report-Only",
        "description": "Allows web developers to experiment with policies by monitoring, but not enforcing, their effects. These violation reports consist of JSON documents sent via an HTTP POST request to the specified URI.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Content-Security-Policy-Report-Only",
    },
    {
        "section": "Security",
        "name": "Permissions-Policy",
        "description": "Provides a mechanism to allow and deny the use of browser features in a website's own frame, and in <iframe>s that it embeds.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Permissions-Policy",
    },
    {
        "section": "Security",
        "name": "Strict-Transport-Security",
        "description": "Force communication using HTTPS instead of HTTP.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Strict-Transport-Security",
    },
    {
        "section": "Security",
        "name": "Upgrade-Insecure-Requests",
        "description": "Sends a signal to the server expressing the client's preference for an encrypted and authenticated response, and that it can successfully handle the upgrade-insecure-requests directive.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Upgrade-Insecure-Requests",
    },
    {
        "section": "Security",
        "name": "X-Content-Type-Options",
        "description": "Disables MIME sniffing and forces browser to use the type given in Content-Type.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/X-Content-Type-Options",
    },
    {
        "section": "Security",
        "name": "X-Frame-Options",
        "description": "Indicates whether a browser should be allowed to render a page in a <frame>, <iframe>, <embed> or <object>.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/X-Frame-Options",
    },
    {
        "section": "Security",
        "name": "X-XSS-Protection",
        "description": "Enables cross-site scripting filtering.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/X-XSS-Protection",
    },
    {
        "section": "Server-sent events",
        "name": "NEL",
        "description": "Defines a mechanism that enables developers to declare a network error reporting policy.",
        "experimental": True,
        "url": "/en-US/docs/Web/HTTP/Headers/NEL",
    },
    {
        "section": "Transfer coding",
        "name": "Transfer-Encoding",
        "description": "Specifies the form of encoding used to safely transfer the resource to the user.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Transfer-Encoding",
    },
    {
        "section": "Transfer coding",
        "name": "TE",
        "description": "Specifies the transfer encodings the user agent is willing to accept.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/TE",
    },
    {
        "section": "Transfer coding",
        "name": "Trailer",
        "description": "Allows the sender to include additional fields at the end of chunked message.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Trailer",
    },
    {
        "section": "Other",
        "name": "Alt-Svc",
        "description": "Used to list alternate ways to reach this service.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Alt-Svc",
    },
    {
        "section": "Other",
        "name": "Alt-Used",
        "description": "Used to identify the alternative service in use.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Alt-Used",
    },
    {
        "section": "Other",
        "name": "Date",
        "description": "Contains the date and time at which the message was originated.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Date",
    },
    {
        "section": "Other",
        "name": "Link",
        "description": "This entity-header field provides a means for serializing one or more links in HTTP headers. It is semantically equivalent to the HTML <link> element.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Link",
    },
    {
        "section": "Other",
        "name": "Retry-After",
        "description": "Indicates how long the user agent should wait before making a follow-up request.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Retry-After",
    },
    {
        "section": "Other",
        "name": "Server-Timing",
        "description": "Communicates one or more metrics and descriptions for the given request-response cycle.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Server-Timing",
    },
    {
        "section": "Other",
        "name": "SourceMap",
        "description": "Links generated code to a source map.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/SourceMap",
    },
    {
        "section": "Other",
        "name": "Upgrade",
        "description": "This HTTP/1.1 (only) header can be used to upgrade an already established client/server connection to a different protocol (over the same transport protocol). For example, it can be used by a client to upgrade a connection from HTTP 1.1 to HTTP 2.0, or an HTTP or HTTPS connection into a WebSocket.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Upgrade",
    },
    {
        "section": "Experimental",
        "name": "Speculation-Rules",
        "description": "Provides a list of URLs pointing to text resources containing speculation rule JSON definitions. When the response is an HTML document, these rules will be added to the document's speculation rule set.",
        "experimental": True,
        "url": "/en-US/docs/Web/HTTP/Headers/Speculation-Rules",
    },
    {
        "section": "Experimental",
        "name": "Supports-Loading-Mode",
        "description": "Set by a navigation target to opt-in to using various higher-risk loading modes. For example, cross-origin, same-site prerendering requires a Supports-Loading-Mode value of credentialed-prerender.",
        "experimental": True,
        "url": "/en-US/docs/Web/HTTP/Headers/Supports-Loading-Mode",
    },
    {
        "section": "Non-standard headers",
        "name": "X-Forwarded-For",
        "description": "Identifies the originating IP addresses of a client connecting to a web server through an HTTP proxy or a load balancer.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/X-Forwarded-For",
    },
    {
        "section": "Non-standard headers",
        "name": "X-Forwarded-Host",
        "description": "Identifies the original host requested that a client used to connect to your proxy or load balancer.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/X-Forwarded-Host",
    },
    {
        "section": "Non-standard headers",
        "name": "X-Forwarded-Proto",
        "description": "Identifies the protocol (HTTP or HTTPS) that a client used to connect to your proxy or load balancer.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/X-Forwarded-Proto",
    },
    {
        "section": "Non-standard headers",
        "name": "X-DNS-Prefetch-Control",
        "description": "Controls DNS prefetching, a feature by which browsers proactively perform domain name resolution on both links that the user may choose to follow as well as URLs for items referenced by the document, including images, CSS, JavaScript, and so forth.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/X-DNS-Prefetch-Control",
    },
    {
        "section": "Deprecated headers",
        "name": "Pragma",
        "description": "Implementation-specific header that may have various effects anywhere along the request-response chain. Used for backwards compatibility with HTTP/1.0 caches where the Cache-Control header is not yet present.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Pragma",
    },
    {
        "section": "Deprecated headers",
        "name": "Warning",
        "description": "General warning information about possible problems.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Warning",
    },
    {
        "section": "Redirects",
        "name": "Refresh",
        "description": 'Directs the browser to reload the page or redirect to another. Takes the same value as the meta element with http-equiv="refresh".',
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Refresh",
    },
    {
        "section": "Security",
        "name": "X-Permitted-Cross-Domain-Policies",
        "description": "Specifies if a cross-domain policy file (crossdomain.xml) is allowed. The file may define a policy to grant clients, such as Adobe's Flash Player (now obsolete), Adobe Acrobat, Microsoft Silverlight (now obsolete), or Apache Flex, permission to handle data across domains that would otherwise be restricted due to the Same-Origin Policy. See the Cross-domain Policy File Specification for more information.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/X-Permitted-Cross-Domain-Policies",
    },
    {
        "section": "Security",
        "name": "X-Powered-By",
        "description": "May be set by hosting environments or other frameworks and contains information about them while not providing any usefulness to the application or its visitors. Unset this header to avoid exposing potential vulnerabilities.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/X-Powered-By",
    },
    {
        "section": "Server-sent events",
        "name": "Report-To",
        "description": "Used to specify a server endpoint for the browser to send warning and error reports to.",
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/Report-To",
    },
    {
        "section": "Experimental headers",
        "name": "Origin-Isolation",
        "description": "Provides a mechanism to allow web applications to isolate their origins.",
        "experimental": True,
        "url": "/en-US/docs/Web/HTTP/Headers/Origin-Isolation",
    },
    {
        "section": "Experimental headers",
        "name": "Accept-Push-Policy",
        "description": "A client can express the desired push policy for a request by sending an Accept-Push-Policy header field in the request.",
        "experimental": True,
        "url": "/en-US/docs/Web/HTTP/Headers/Accept-Push-Policy",
    },
    {
        "section": "Experimental headers",
        "name": "Accept-Signature",
        "description": "A client can send the Accept-Signature header field to indicate intention to take advantage of any available signatures and to indicate what kinds of signatures it supports.",
        "experimental": True,
        "url": "/en-US/docs/Web/HTTP/Headers/Accept-Signature",
    },
    {
        "section": "Experimental headers",
        "name": "Early-Data",
        "description": "Indicates that the request has been conveyed in TLS early data.",
        "experimental": True,
        "url": "/en-US/docs/Web/HTTP/Headers/Early-Data",
    },
    {
        "section": "Experimental headers",
        "name": "Push-Policy",
        "description": "A Push-Policy defines the server behavior regarding push when processing a request.",
        "experimental": True,
        "url": "/en-US/docs/Web/HTTP/Headers/Push-Policy",
    },
    {
        "section": "Experimental headers",
        "name": "Signature",
        "description": "The Signature header field conveys a list of signatures for an exchange, each one accompanied by information about how to determine the authority of and refresh that signature.",
        "experimental": True,
        "url": "/en-US/docs/Web/HTTP/Headers/Signature",
    },
    {
        "section": "Experimental headers",
        "name": "Signed-Headers",
        "description": "The Signed-Headers header field identifies an ordered list of response header fields to include in a signature.",
        "experimental": True,
        "url": "/en-US/docs/Web/HTTP/Headers/Signed-Headers",
    },
    {
        "section": "Privacy",
        "name": "Sec-GPC",
        "description": "Indicates whether the user consents to a website or service selling or sharing their personal information with third parties.",
        "experimental": True,
        "url": "/en-US/docs/Web/HTTP/Headers/Sec-GPC",
    },
    {
        "section": "Non-standard headers",
        "name": "X-Robots-Tag",
        "description": 'The X-Robots-Tag HTTP header is used to indicate how a web page is to be indexed within public search engine results. The header is effectively equivalent to <meta name="robots" content="…">.',
        "experimental": False,
        "url": "/en-US/docs/Web/HTTP/Headers/X-Robots-Tag",
    },
]


request_header_names = {
    "Authorization",
    "Proxy-Authorization",
    "Cache-Control",
    "Clear-Site-Data",
    "If-Match",
    "If-None-Match",
    "If-Modified-Since",
    "If-Unmodified-Since",
    "Connection",
    "Keep-Alive",
    "Accept",
    "Accept-Encoding",
    "Accept-Language",
    "Expect",
    "Max-Forwards",
    "Cookie",
    "Access-Control-Request-Headers",
    "Access-Control-Request-Method",
    "Origin",
    "Timing-Allow-Origin",
    "Content-Length",
    "Content-Type",
    "Content-Encoding",
    "Content-Language",
    "Content-Location",
    "Forwarded",
    "Via",
    "From",
    "Host",
    "Referer",
    "Referrer-Policy",
    "User-Agent",
    "Range",
    "If-Range",
    "Upgrade-Insecure-Requests",
    "Transfer-Encoding",
    "TE",
    "Trailer",
    "Alt-Used",
    "Link",
    "Date",
    "Origin-Isolation",
    "Accept-Push-Policy",
    "Accept-Signature",
    "Early-Data",
    "Signature",
    "Signed-Headers",
    "Sec-GPC",
    "X-Forwarded-For",
    "X-Forwarded-Host",
    "X-Forwarded-Proto",
    "X-DNS-Prefetch-Control",
    "Pragma",
    "Warning",
}

REQUEST_HEADERS = [
    header for header in headers if header["name"] in request_header_names
]


_VARIABLES_PATTERN = re.compile(
    r"\$(?:([a-zA-Z_][a-zA-Z0-9_]*)|{([a-zA-Z_][a-zA-Z0-9_]*)})"
)


def get_variables() -> dict[str, str | None]:
    return {name: "" for name in request_header_names}


@lru_cache()
def find_variables(template_str: str) -> list[tuple[str, int, int]]:
    return [
        (m.group(1) or m.group(2), m.start(), m.end())
        for m in re.finditer(_VARIABLES_PATTERN, template_str)
    ]


@lru_cache()
def is_cursor_within_variable(cursor: int, text: str) -> bool:
    if not text or cursor < 0 or cursor > len(text):
        return False

    # Check for ${var} syntax
    start = text.rfind("${", 0, cursor)
    if start != -1:
        end = text.find("}", start)
        if end != -1 and start < cursor <= end:
            return True

    # Check for $ followed by { (cursor between $ and {)
    if (
        cursor > 0
        and text[cursor - 1] == "$"
        and cursor < len(text)
        and text[cursor] == "{"
    ):
        return True

    # Check for $var syntax
    last_dollar = text.rfind("$", 0, cursor)
    if last_dollar == -1:
        return False

    # Check if cursor is within a valid variable name
    for i in range(last_dollar + 1, len(text)):
        if i >= cursor:
            return True
        if not (text[i].isalnum() or text[i] == "_"):
            return False

    return True


@lru_cache()
def find_variable_start(cursor: int, text: str) -> int:
    if not text:
        return cursor

    # Check for ${var} syntax
    start = text.rfind("${", 0, cursor)
    if start != -1 and start < cursor <= text.find("}", start):
        return start

    # Check for $ followed by { (cursor between $ and {)
    if (
        cursor > 0
        and text[cursor - 1] == "$"
        and cursor < len(text)
        and text[cursor] == "{"
    ):
        return cursor - 1

    # Check for $var syntax
    for i in range(cursor - 1, -1, -1):
        if text[i] == "$":
            if i + 1 < len(text) and text[i + 1] == "{":
                return i
            if all(c.isalnum() or c == "_" for c in text[i + 1 : cursor]):
                return i

    return cursor  # No valid variable start found


@lru_cache()
def find_variable_end(cursor: int, text: str) -> int:
    if not text:
        return cursor

    # Check for ${var} syntax
    start = text.rfind("${", 0, cursor)
    if start != -1:
        end = text.find("}", start)
        if end != -1 and start < cursor <= end + 1:  # Include the closing brace
            return end + 1

    # Check for $ followed by { (cursor between $ and {)
    if (
        cursor > 0
        and text[cursor - 1] == "$"
        and cursor < len(text)
        and text[cursor] == "{"
    ):
        end = text.find("}", cursor)
        if end != -1:
            return end + 1

    # Check for $var syntax
    for i in range(cursor, len(text)):
        if not (text[i].isalnum() or text[i] == "_"):
            return i

    return len(text)


@lru_cache()
def get_variable_at_cursor(cursor_column: int, text: str) -> str | None:
    if not is_cursor_within_variable(cursor_column, text):
        return None

    start = find_variable_start(cursor_column, text)
    end = find_variable_end(cursor_column, text)

    return text[start:end]


def extract_variable_name(variable_text: str) -> str:
    """
    Extract the variable name from a variable reference.

    Args:
        variable_text: The text containing the variable reference.

    Returns:
        str: The extracted variable name.

    Examples:
        >>> extract_variable_name("$var")
        'var'
        >>> extract_variable_name("${MY_VAR}")
        'MY_VAR'
    """
    if variable_text.startswith("${") and variable_text.endswith("}"):
        return variable_text[2:-1]
    elif variable_text.startswith("$"):
        return variable_text[1:]
    return variable_text  # Return as-is if it doesn't match expected formats


class SubstitutionError(Exception):
    """Raised when the user refers to a variable that doesn't exist."""


class VariableAutoComplete(AutoComplete):
    def __init__(
        self,
        target: Input | TextArea | str,
        candidates: list[DropdownItem] | Callable[[TargetState], list[DropdownItem]],
        variable_candidates: list[DropdownItem]
        | Callable[[TargetState], list[DropdownItem]]
        | None = None,
        matcher_factory: MatcherFactoryType | None = None,
        prevent_default_enter: bool = True,
        prevent_default_tab: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            target,
            candidates,
            matcher_factory,
            None,
            self._search_string,
            prevent_default_enter,
            prevent_default_tab,
            name,
            id,
            classes,
            disabled,
        )
        if variable_candidates is None:
            variable_candidates = [
                DropdownItem(main=f"${variable}") for variable in get_variables()
            ]
        self.variable_candidates = variable_candidates

    def get_candidates(self, target_state: TargetState) -> list[DropdownItem]:
        cursor = target_state.selection.end[1]
        text = target_state.text
        if is_cursor_within_variable(cursor, text):
            return self.get_variable_candidates(target_state)
        else:
            return super().get_candidates(target_state)

    def _search_string(self, target_state: TargetState) -> str:
        cursor_row, cursor_column = target_state.selection.end
        text = target_state.text
        target = self.target

        if isinstance(target, TextArea):
            text = target.document.get_line(cursor_row)

        if is_cursor_within_variable(cursor_column, text):
            variable_at_cursor = get_variable_at_cursor(cursor_column, text)
            return variable_at_cursor or ""
        else:
            if isinstance(target, TextArea):
                # Use the word behind the cursor as a search string
                start, end = self.get_text_area_word_bounds_before_cursor(self.target)
                search_string = target.get_text_range(start, end)
                print("DEFAULT search_string", search_string)
                return search_string
            else:
                # Use the entire Input as a search string
                return target.value

    def get_variable_candidates(self, target_state: TargetState) -> list[DropdownItem]:
        candidates = self.variable_candidates
        return candidates(target_state) if callable(candidates) else candidates


class Version2(App[None]):
    CSS = "Vertical { height: 100; } TextArea { height: 1fr; }"

    BINDINGS = [
        Binding("ctrl+n", "insert_matching_text", "Insert matching text"),
    ]

    def compose(self) -> ComposeResult:
        # input = TextArea()
        with VerticalScroll():
            with Vertical():
                input_one = Input(id="my-input")
                input_one.focus()
                yield input_one
                yield Input()
                yield TextArea.code_editor()
        yield Footer()

    def on_mount(self) -> None:
        # If we mount it like this it ensures it's on the screen.
        # This might be easier to explain.
        # OR we could provide a utility function that does this for us.

        items: list[DropdownItem] = []
        for header in REQUEST_HEADERS:
            style = "yellow" if header["experimental"] else ""
            items.append(DropdownItem(Text(header["name"], style=style)))

        self.screen.mount(
            VariableAutoComplete(
                target=self.query_one("#my-input", Input),
                candidates=[
                    DropdownItem(main=header) for header in request_header_names
                ],
                variable_candidates=[
                    DropdownItem(main=f"${var}") for var in os.environ.keys()
                ],
            )
        )

        self.screen.mount(
            VariableAutoComplete(
                target=self.query_one(TextArea),
                candidates=[
                    DropdownItem(main=header) for header in request_header_names
                ],
                variable_candidates=[
                    DropdownItem(main=f"${var}") for var in os.environ.keys()
                ],
            )
        )

    def action_insert_matching_text(self) -> None:
        self.query_one("#my-input", Input).value = "Authorization"


if __name__ == "__main__":
    app = Version2()
    app.run()
